/*
types taken from highlight.js/types
*/

export type HLJSApi = PublicApi & ModesAPI;

export type LanguageFn = (hljs: HLJSApi) => Language;

interface PublicApi {
  highlight: (
    codeOrLanguageName: string,
    optionsOrCode: string | HighlightOptions,
    ignoreIllegals?: boolean
  ) => HighlightResult;
  highlightAuto: (code: string, languageSubset?: string[]) => AutoHighlightResult;
  highlightBlock: (element: HTMLElement) => void;
  highlightElement: (element: HTMLElement) => void;
  configure: (options: Partial<HLJSOptions>) => void;
  initHighlighting: () => void;
  initHighlightingOnLoad: () => void;
  highlightAll: () => void;
  registerLanguage: (languageName: string, language: LanguageFn) => void;
  unregisterLanguage: (languageName: string) => void;
  listLanguages: () => string[];
  registerAliases: (
    aliasList: string | string[],
    { languageName }: { languageName: string }
  ) => void;
  getLanguage: (languageName: string) => Language | undefined;
  autoDetection: (languageName: string) => boolean;
  inherit: <T>(original: T, ...args: Record<string, any>[]) => T;
  addPlugin: (plugin: HLJSPlugin) => void;
  debugMode: () => void;
  safeMode: () => void;
  versionString: string;
  vuePlugin: () => VuePlugin;
}
interface VuePlugin {
  install: (vue: any) => void;
}
type HLJSPlugin = {
  'after:highlight'?: (result: HighlightResult) => void;
  'before:highlight'?: (context: BeforeHighlightContext) => void;
  'after:highlightElement'?: (data: { el: Element; result: HighlightResult; text: string }) => void;
  'before:highlightElement'?: (data: { el: Element; language: string }) => void;
  // TODO: Old API, remove with v12
  'after:highlightBlock'?: (data: {
    block: Element;
    result: HighlightResult;
    text: string;
  }) => void;
  'before:highlightBlock'?: (data: { block: Element; language: string }) => void;
};
interface HighlightOptions {
  language: string;
  ignoreIllegals?: boolean;
}
type BeforeHighlightContext = {
  code: string;
  language: string;
  result?: HighlightResult;
};
interface HighlightResult {
  code?: string;
  relevance: number;
  value: string;
  language?: string;
  illegal: boolean;
  errorRaised?: Error;
  // * for auto-highlight
  secondBest?: Omit<HighlightResult, 'second_best'>;
  // private
  _illegalBy?: illegalData;
  _emitter: Emitter;
  _top?: Language | CompiledMode;
}
interface AutoHighlightResult extends HighlightResult {}
interface illegalData {
  message: string;
  context: string;
  index: number;
  resultSoFar: string;
  mode: CompiledMode;
}
interface HLJSOptions {
  noHighlightRe: RegExp;
  languageDetectRe: RegExp;
  classPrefix: string;
  cssSelector: string;
  languages?: string[];
  __emitter: EmitterConstructor;
  ignoreUnescapedHTML?: boolean;
}
type CompiledMode = Omit<Mode, 'contains'> & {
  contains: CompiledMode[];
  keywords: KeywordDict;
  data: Record<string, any>;
  terminatorEnd: string;
  keywordPatternRe: RegExp;
  beginRe: RegExp;
  endRe: RegExp;
  illegalRe: RegExp;
  matcher: any;
  isCompiled: true;
  starts?: CompiledMode;
  parent?: CompiledMode;
  beginScope?: Record<number, string> & {
    _emit?: Record<number, boolean>;
    _multi?: boolean;
    _wrap?: string;
  };
  endScope?: Record<number, string> & {
    _emit?: Record<number, boolean>;
    _multi?: boolean;
    _wrap?: string;
  };
};
interface EmitterConstructor {
  new (opts: any): Emitter;
}
interface Emitter {
  addKeyword(text: string, kind: string): void;
  addText(text: string): void;
  toHTML(): string;
  finalize(): void;
  closeAllNodes(): void;
  openNode(kind: string): void;
  closeNode(): void;
  addSublanguage(emitter: Emitter, subLanguageName: string): void;
}
interface ModesAPI {
  SHEBANG: (mode?: Partial<Mode> & { binary?: string | RegExp }) => Mode;
  BACKSLASH_ESCAPE: Mode;
  QUOTE_STRING_MODE: Mode;
  APOS_STRING_MODE: Mode;
  PHRASAL_WORDS_MODE: Mode;
  COMMENT: (begin: string | RegExp, end: string | RegExp, modeOpts?: Mode | {}) => Mode;
  C_LINE_COMMENT_MODE: Mode;
  C_BLOCK_COMMENT_MODE: Mode;
  HASH_COMMENT_MODE: Mode;
  NUMBER_MODE: Mode;
  C_NUMBER_MODE: Mode;
  BINARY_NUMBER_MODE: Mode;
  REGEXP_MODE: Mode;
  TITLE_MODE: Mode;
  UNDERSCORE_TITLE_MODE: Mode;
  METHOD_GUARD: Mode;
  END_SAME_AS_BEGIN: (mode: Mode) => Mode;
  // built in regex
  IDENT_RE: string;
  UNDERSCORE_IDENT_RE: string;
  MATCH_NOTHING_RE: string;
  NUMBER_RE: string;
  C_NUMBER_RE: string;
  BINARY_NUMBER_RE: string;
  RE_STARTERS_RE: string;
}

type KeywordData = [string, number];
type KeywordDict = Record<string, KeywordData>;
interface Mode extends ModeCallbacks, ModeDetails {}
export type Language = LanguageDetail & Partial<Mode>;
type CompilerExt = (mode: Mode, parent: Mode | Language | null) => void;
interface LanguageDetail {
  name?: string;
  rawDefinition?: () => Language;
  aliases?: string[];
  disableAutodetect?: boolean;
  contains: Mode[];
  case_insensitive?: boolean;
  keywords?: Record<string, any> | string;
  isCompiled?: boolean;
  exports?: any;
  classNameAliases?: Record<string, string>;
  compilerExtensions?: CompilerExt[];
  supersetOf?: string;
}
type ModeCallback = (match: RegExpMatchArray, response: CallbackResponse) => void;
interface ModeCallbacks {
  'on:end'?: Function;
  'on:begin'?: ModeCallback;
}
interface ModeDetails {
  begin?: RegExp | string | (RegExp | string)[];
  match?: RegExp | string | (RegExp | string)[];
  end?: RegExp | string | (RegExp | string)[];
  // deprecated in favor of `scope`
  className?: string;
  scope?: string | Record<number, string>;
  beginScope?: string | Record<number, string>;
  endScope?: string | Record<number, string>;
  contains?: ('self' | Mode)[];
  endsParent?: boolean;
  endsWithParent?: boolean;
  endSameAsBegin?: boolean;
  skip?: boolean;
  excludeBegin?: boolean;
  excludeEnd?: boolean;
  returnBegin?: boolean;
  returnEnd?: boolean;
  __beforeBegin?: Function;
  parent?: Mode;
  starts?: Mode;
  lexemes?: string | RegExp;
  keywords?: Record<string, any> | string;
  beginKeywords?: string;
  relevance?: number;
  illegal?: string | RegExp | Array<string | RegExp>;
  variants?: Mode[];
  cachedVariants?: Mode[];
  // parsed
  subLanguage?: string | string[];
  isCompiled?: boolean;
  label?: string;
}
interface CallbackResponse {
  data: Record<string, any>;
  ignoreMatch: () => void;
  isMatchIgnored: boolean;
}
